/*
 *  Copyright (c) 2015, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.wso2.maven.p2.profile;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.Component;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.MavenProject;
import org.eclipse.sisu.equinox.launching.internal.P2ApplicationLauncher;
import org.eclipse.tycho.model.ProductConfiguration;
import org.wso2.maven.p2.beans.product.config.ProductFileConfig;
import org.wso2.maven.p2.utils.FileManagementUtil;
import org.wso2.maven.p2.utils.P2ApplicationLaunchManager;
import org.wso2.maven.p2.utils.P2Constants;
import org.wso2.maven.p2.utils.ProductFileUtils;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.nio.file.Paths;
import javax.xml.bind.JAXBException;

/**
 * Mojo responsible for generating a runtime.
 *
 * @since 2.0.0
 */
@Mojo(name = "generate-runtime")
public class GenerateProfileMojo extends AbstractMojo {

    @Parameter(required = true, defaultValue = "${project}")
    protected MavenProject project;
    /**
     * Metadata repository name.
     */
    @Parameter
    private URL repositoryURL;

    /**
     * The product configuration, a .product file. This file manages all aspects
     * of a product definition from its constituent plug-ins to configuration
     * files to branding.
     */
    @Parameter(defaultValue = "${productConfiguration}")
    private File productConfigurationFile;

    /**
     * Bean to capture build-time product file customizations
     * from pom level configurations.
     */
    @Parameter
    private ProductFileConfig productFileConfig = new ProductFileConfig();

    @Parameter
    private URL targetPath;

    /**
     * The new runtime to be created during p2 Director install &
     * the default runtime for the the application which is set in config.ini.
     */
    @Parameter(defaultValue = "${profile}")
    private String runtime;


    @Component
    private P2ApplicationLauncher launcher;

    /**
     * Kill the forked test process after a certain number of seconds. If set to 0, wait forever for
     * the process, never timing out.
     */
    @Parameter(defaultValue = "${p2.timeout}")
    private int forkedProcessTimeoutInSeconds;

    /**
     * Overridden method of AbstractMojo class. This is picked up by the maven runtime for execution.
     *
     * @throws MojoExecutionException throws when any runtime exception occurs. i.e: fail to read write file, fail to
     *                                parse a configuration xml
     * @throws MojoFailureException   throws when the tool breaks for any configuration issues
     */
    public void execute() throws MojoExecutionException, MojoFailureException {
        try {
            if (runtime == null) {
                runtime = P2Constants.DEFAULT_PROFILE_ID;
            }
            deployRepository();
            //updating runtime's config.ini p2.data.area property using relative path
            File profileConfigIni = FileManagementUtil.getProfileConfigIniFile(targetPath.getPath(), runtime);
            FileManagementUtil.changeConfigIniProperty(profileConfigIni, "eclipse.p2.data.area",
                    P2Constants.P2_DIRECTORY, this.getLog());
        } catch (IOException e) {
            throw new MojoExecutionException(e.getMessage(), e);
        }
    }

    private void deployRepository() throws MojoFailureException, MojoExecutionException, IOException {
        ProductConfiguration productConfiguration;
        if (productConfigurationFile != null) {
            productConfiguration = ProductConfiguration.read(productConfigurationFile);
        } else {
            // going ahead with a dynamically generated product file during build-time
            File productFile = Paths.get(ProductFileUtils.getProductFilePath(project)).toFile();
            if (!productFile.exists()) {
                try {
                    ProductFileUtils.generateProductFile(productFileConfig, project);
                } catch (JAXBException e) {
                    throw new MojoExecutionException("Cannot proceed as there is an error in " +
                            "writing configurations to product file.", e);
                }
            }
            productConfiguration = ProductConfiguration.read(productFile);
        }
        P2ApplicationLaunchManager p2LaunchManager = new P2ApplicationLaunchManager(launcher);
        p2LaunchManager.setWorkingDirectory(project.getBasedir());
        p2LaunchManager.setApplicationName("org.eclipse.equinox.p2.director");
        p2LaunchManager.addGenerateProfileArguments(repositoryURL, productConfiguration.getId(), runtime, targetPath);
        p2LaunchManager.performAction(forkedProcessTimeoutInSeconds);
    }
}
